# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import cgi, gtk, modules, pynotify, os.path, string

from gui           import infoMsgBox
from tools         import consts, prefs
from modules       import ThreadedModule
from gettext       import gettext as _
from gui.modWindow import ModWindow


# Module configuration
MOD_NAME            = _('Desktop Notification')
MOD_DESC            = _('Display a desktop notification on track change')
MOD_IS_MANDATORY    = False
MOD_IS_CONFIGURABLE = True


# Default preferences
PREFS_DEFAULT_BODY    = 'by {artist}\nfrom {album}'
PREFS_DEFAULT_TITLE   = '<b>{title}</b>  [{duration_str}]'
PREFS_DEFAULT_TIMEOUT = 10


# Fields that may be used to configure this module
FIELDS = (
            ( 'track'       , _('Track number')                      ),
            ( 'title'       , _('Track title')                       ),
            ( 'artist'      , _('Track artist')                      ),
            ( 'album'       , _('Track album')                       ),
            ( 'duration_sec', _('Duration in seconds (e.g., 194)')   ),
            ( 'duration_str', _('Duration as a string (e.g., 3:14)') ),
            ( 'path'        , _('Full path to the file')             )
         )


class DesktopNotification(ThreadedModule):


    def __init__(self, wTree):
        """ Constructor """
        ThreadedModule.__init__(self)
        # Widgets
        self.notif  = None
        self.window = None
        # It seems that this call must be done in the GTK main loop
        pynotify.init(consts.appNameShort)
        modules.register(self, [modules.MSG_EVT_NEW_TRACK])


    def __format(self, string, track):
        """ Replace the special fields in the given string by their corresponding value """
        result = string

        result = result.replace( '{path}',         cgi.escape(track.getPath())   )
        result = result.replace( '{album}',        cgi.escape(track.getAlbum())  )
        result = result.replace( '{track}',        str(track.getNumber())        )
        result = result.replace( '{title}',        cgi.escape(track.getTitle())  )
        result = result.replace( '{artist}',       cgi.escape(track.getArtist()) )
        result = result.replace( '{duration_sec}', str(track.getLengthSec())     )
        result = result.replace( '{duration_str}', track.getLengthStr()          )

        return result


    def configure(self, parent):
        """ Show the configuration window """
        if self.window is None:
            self.window      = ModWindow('DesktopNotification.glade', 'vbox1', __name__, MOD_NAME, 355, 345)
            self.btnOk       = self.window.getWidget('btn-ok')
            self.txtBody     = self.window.getWidget('txt-body')
            self.txtTitle    = self.window.getWidget('txt-title')
            self.spnDuration = self.window.getWidget('spn-duration')
            # GTK handlers
            self.window.getWidget('btn-help').connect('clicked', self.onHelp)
            self.window.getWidget('btn-cancel').connect('clicked', lambda btn: self.window.hide())
            self.btnOk.connect('clicked', self.onOk)

        # Fill current preferences
        if not self.window.isVisible():
            self.txtTitle.set_text(prefs.get(__name__, 'title', PREFS_DEFAULT_TITLE))
            self.txtBody.get_buffer().set_text(prefs.get(__name__, 'body', PREFS_DEFAULT_BODY))
            self.spnDuration.set_value(prefs.get(__name__, 'timeout', PREFS_DEFAULT_TIMEOUT))
            self.btnOk.grab_focus()

        self.window.show()


    # --== GTK handlers ==--


    def onHelp(self, btn):
        """ Display a small help message box """
        helpMsg  = _('Any field of the form <b>{<tt>field</tt>}</b> will be replaced by the corresponding value. Available fields are:\n\n')
        helpMsg += '\n'.join(['<tt><b>%s</b></tt> %s' % (string.ljust(field, 15), desc) for (field, desc) in FIELDS])
        infoMsgBox(self.window, helpMsg)


    def onOk(self, btn):
        """ Save the new preferences """
        prefs.set(__name__, 'title', self.txtTitle.get_text())
        prefs.set(__name__, 'timeout', int(self.spnDuration.get_value()))
        (start, end) = self.txtBody.get_buffer().get_bounds()
        prefs.set(__name__, 'body', self.txtBody.get_buffer().get_text(start, end))
        self.window.hide()


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ Handle messages sent to this module """
        if msg == modules.MSG_EVT_NEW_TRACK:
            body  = self.__format(prefs.get(__name__, 'body', PREFS_DEFAULT_BODY), params['track'])
            title = '</b>%s<b>' % self.__format(prefs.get(__name__, 'title', PREFS_DEFAULT_TITLE), params['track'])

            # Try to use the app's icon
            icon = os.path.join(consts.dirPix, 'decibel-audio-player-64.png')
            if not os.path.exists(icon):
                icon = gtk.STOCK_DIALOG_INFO

            if self.notif is None: self.notif = pynotify.Notification(title, body, icon)
            else:                  self.notif.update(title, body, icon)

            self.notif.set_timeout(prefs.get(__name__, 'timeout', PREFS_DEFAULT_TIMEOUT) * 1000)
            self.notif.show()
