# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, gtk.glade, media, modules, os, tools, urllib

from gtk           import gdk
from gui           import fileChooser, infoMsgBox
from tools         import consts, prefs
from media         import playlist
from modules       import Module
from gettext       import gettext as _
from gui.modWindow import ModWindow


# Module configuration
MOD_NAME            = _('File Explorer')
MOD_DESC            = _('Navigate through your file system')
MOD_IS_MANDATORY    = True
MOD_IS_CONFIGURABLE = True


# Default preferences
PREFS_DEFAULT_MEDIA_FOLDERS = {_('Home'): consts.dirUsr}    # List of media folders that can be used as roots for the file explorer


# The format of a row in the treeview
(
    ROW_NAME,      # Item name
    ROW_TYPE,      # The type of the item (e.g., directory, file)
    ROW_PIXBUF,    # Item icon
    ROW_FULLPATH   # The full path to the item
) = range(4)


# The format of a row in the preferences' list (media folders)
(
    ROW_FOLDER_NAME,
    ROW_FOLDER_PATH
) = range(2)


# The possible types for an item
(
    TYPE_DIR,   # A directory
    TYPE_PLIST, # A playlist
    TYPE_FILE,  # A media file
    TYPE_NONE   # A fake item, used to display a '+' in front of a directory when needed
) = range(4)


# Data associated with a media folder
(
    MF_PIXBUF,    # The pixbuf to display in the explorer combo box
    MF_FULLPATH   # The full path to the folder
) = range(2)


class FileExplorer(Module):
    """
        This explorer lets the user browse the disk from a given root directory (e.g., ~/, /)
        It uses a 'lazy' treeview (i.e., it is populated only when needed)
    """


    def __init__(self, wTree):
        """ Constructor """
        Module.__init__(self)
        self.currRoot        = None
        self.mediaFolders    = prefs.get(__name__, 'media-folders', PREFS_DEFAULT_MEDIA_FOLDERS)
        self.showHiddenFiles = False
        # Widgets
        self.tree      = gtk.TreeView()
        self.scrolled  = gtk.ScrolledWindow()
        self.popupMenu = None
        self.cfgWindow = None
        # Icons
        self.dirIcon      = self.tree.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU)
        self.mediaIcon    = self.tree.render_icon(gtk.STOCK_FILE,      gtk.ICON_SIZE_MENU).copy()  # We need a copy to modify it
        self.dirMediaIcon = self.tree.render_icon(gtk.STOCK_DIRECTORY, gtk.ICON_SIZE_MENU).copy()  # We need a copy to modify it

        cdrom = self.tree.render_icon(gtk.STOCK_CDROM, gtk.ICON_SIZE_MENU)
        cdrom.composite(self.mediaIcon,    5, 5, 11, 11, 5, 5, 0.6875, 0.6875, gdk.INTERP_HYPER, 255)
        cdrom.composite(self.dirMediaIcon, 5, 5, 11, 11, 5, 5, 0.6875, 0.6875, gdk.INTERP_HYPER, 255)
        # Configure the widgets
        self.scrolled.add(self.tree)
        self.scrolled.set_shadow_type(gtk.SHADOW_IN)
        self.scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.scrolled.show()

        imgRdr = gtk.CellRendererPixbuf()
        txtRdr = gtk.CellRendererText()
        column = gtk.TreeViewColumn()
        self.tree.append_column(column)
        self.tree.set_headers_visible(False)
        column.pack_start(imgRdr, False)
        column.add_attribute(imgRdr, 'pixbuf', ROW_PIXBUF)
        column.pack_start(txtRdr, False)
        column.add_attribute(txtRdr, 'text', ROW_NAME)
        column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

        self.store = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_INT, gdk.Pixbuf, gobject.TYPE_STRING)
        self.tree.set_model(self.store)
        self.tree.enable_model_drag_source(gdk.BUTTON1_MASK, [('text/uri-list', gtk.TARGET_SAME_APP, 0)], gdk.ACTION_DEFAULT | gdk.ACTION_COPY)
        self.tree.show()
        # Messages handled by this module
        modules.register(self, [modules.MSG_EVT_APP_STARTED, modules.MSG_EVT_EXPLORER_CHANGED])
        # GTK handlers
        self.tree.connect('row-expanded',       self.onRowExpanded)
        self.tree.connect('drag-data-get',      self.onDragDataGet)
        self.tree.connect('button-press-event', self.onMouse)


    def updateChildren(self, parentPath, firstChild):
        """ Update 3 children, starting from firstChild, and then repost the same task if there are some children left """
        if parentPath is None: parent = None
        else:                  parent = self.store.get_iter(parentPath)

        self.tree.freeze_child_notify()
        for childId in xrange(firstChild, firstChild+3):

            # Directories are always placed before other files, so we can stop if we find something else than a directory
            child = self.store.iter_nth_child(parent, childId)
            if child is None or self.store.get_value(child, ROW_TYPE) != TYPE_DIR:
                self.tree.thaw_child_notify()
                return

            childPath     = self.store.get_value(child, ROW_FULLPATH)
            needFakeChild = False

            # Take care of permissions
            if not os.access(childPath, os.R_OK | os.X_OK):
                continue

            for file in os.listdir(childPath):
                if file[0] != '.' or self.showHiddenFiles:
                    path = os.path.join(childPath, file)
                    if os.path.isdir(path):
                        needFakeChild = True
                    elif os.path.isfile(path) and (media.isSupported(file) or playlist.isSupported(file)):
                        needFakeChild = True
                        self.store.set_value(child, ROW_PIXBUF, self.dirMediaIcon)
                        break

            if needFakeChild:
                self.store.append(child, ('', TYPE_NONE, self.dirIcon, ''))
        self.tree.thaw_child_notify()

        gobject.idle_add(self.updateChildren, parentPath, firstChild+3)


    def showPopupMenu(self, button, time):
        """ Show a popup menu """
        if self.popupMenu is None:
            wTree          = gtk.glade.XML(os.path.join(consts.dirRes, 'FileExplorerMenu.glade'), domain=consts.appNameShort)
            self.popupMenu = wTree.get_widget('menu-popup')
            # Connect handlers
            wTree.get_widget('item-add').connect ('activate', lambda widget: self.playSelection(False))
            wTree.get_widget('item-play').connect('activate', lambda widget: self.playSelection(True))
        self.popupMenu.popup(None, None, None, button, time)


    def playSelection(self, replace):
        """ Replace/extend the tracklist """
        path = [self.store.get_value(self.tree.get_selection().get_selected()[1], ROW_FULLPATH)]

        if replace: modules.postMsg(modules.MSG_CMD_TRACKLIST_SET, {'files': path, 'playNow': True})
        else:       modules.postMsg(modules.MSG_CMD_TRACKLIST_ADD, {'files': path})


    def __cmpRows(self, r1, r2):
        """ Used to sort rows """
        return cmp(r1[ROW_NAME].lower(), r2[ROW_NAME].lower())


    def exploreDir(self, dir, parent):
        """ List the content of the given directory and append it to the tree as a child of parent """
        allFiles = [(file, os.path.join(dir, file)) for file in os.listdir(dir) if file[0] != '.' or self.showHiddenFiles]

        directories = [(file, TYPE_DIR,   self.dirIcon,   path) for (file, path) in allFiles if os.path.isdir(path)]
        playlists   = [(file, TYPE_PLIST, self.mediaIcon, path) for (file, path) in allFiles if os.path.isfile(path) and playlist.isSupported(file)]
        mediaFiles  = [(file, TYPE_FILE,  self.mediaIcon, path) for (file, path) in allFiles if os.path.isfile(path) and media.isSupported(file)]

        directories.sort(self.__cmpRows)
        playlists.sort(self.__cmpRows)
        mediaFiles.sort(self.__cmpRows)

        for row in directories: self.store.append(parent, row)
        for row in playlists:   self.store.append(parent, row)
        for row in mediaFiles:  self.store.append(parent, row)

        if len(directories) != 0:
            if parent is None: gobject.idle_add(self.updateChildren, None,                        0)
            else:              gobject.idle_add(self.updateChildren, self.store.get_path(parent), 0)


    # --== GTK handlers ==--


    def onMouse(self, tree, event):
        """ A mouse button has been pressed """
        pathInfo = self.tree.get_path_at_pos(int(event.x), int(event.y))
        if pathInfo is not None:
            path = pathInfo[0]
            icon = self.store.get_value(self.store.get_iter(path), ROW_PIXBUF)
            if event.button == 3 and icon != self.dirIcon:
                self.showPopupMenu(event.button, event.time)
            elif event.button == 1 and event.type == gdk._2BUTTON_PRESS:
                if   icon != self.dirIcon:         self.playSelection(True)
                elif self.tree.row_expanded(path): self.tree.collapse_row(path)
                else:                              self.tree.expand_row(path, False)


    def onRowExpanded(self, tree, iter, path):
        """ Populate the expanded row if needed """
        child = self.store.iter_children(iter)
        if self.store.get_value(child, ROW_TYPE) == TYPE_NONE:
            self.tree.freeze_child_notify()
            self.exploreDir(self.store.get_value(iter, ROW_FULLPATH), iter)
            self.store.remove(child)
            self.tree.thaw_child_notify()


    def onDragDataGet(self, tree, context, selection, info, time):
        """ Provide information about the data being dragged """
        model, iter = tree.get_selection().get_selected()
        url = 'file://' + urllib.pathname2url(model.get_value(iter, ROW_FULLPATH))
        selection.set('text/uri-list', 8, url)


    def onAppStarted(self):
        """ Add all media folders to the explorer Module """
        for (name, path) in self.mediaFolders.iteritems():
            modules.postMsg(modules.MSG_CMD_EXPLORER_ADD, {'modName': __name__, 'expName': name, 'icon': None, 'usrParam': path, 'widget': self.scrolled})


   # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ Handle messages sent to this module """
        if msg == modules.MSG_EVT_EXPLORER_CHANGED and params['modName'] == __name__:
            # Change the root directory of the tree, according to the new selection
            if params['expName'] != self.currRoot:
                self.currRoot = params['expName']
                self.tree.freeze_child_notify()
                self.store.clear()
                self.exploreDir(params['usrParam'], None)
                self.tree.scroll_to_cell(0)
                self.tree.thaw_child_notify()
        elif msg == modules.MSG_EVT_APP_STARTED:
            gobject.idle_add(self.onAppStarted)


    # --== Configuration ==--


    def configure(self, parent):
        """ Show the configuration dialog """
        if self.cfgWindow is None:
            self.cfgWindow    = ModWindow('FileExplorer.glade', 'vbox1', __name__, MOD_NAME, 370, 400)
            self.cfgTree      = self.cfgWindow.getWidget('tree-folders')
            self.cfgDlgAdd    = None
            self.cfgBtnAdd    = self.cfgWindow.getWidget('btn-add')
            self.cfgBtnRemove = self.cfgWindow.getWidget('btn-remove')
            # Add columns to the list
            renderer      = gtk.CellRendererText()
            self.cfgStore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
            self.cfgTree.set_model(self.cfgStore)
            self.cfgTree.insert_column_with_attributes(-1, 'Name', renderer, markup=ROW_FOLDER_NAME)
            self.cfgTree.insert_column_with_attributes(-1, 'Path', renderer, markup=ROW_FOLDER_PATH)
            self.cfgTree.get_column(0).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
            self.cfgTree.get_column(1).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
            # Connect handlers
            self.cfgBtnAdd.connect('clicked', self.onAddMediaFolder)
            self.cfgBtnRemove.connect('clicked', self.onRemoveMediaFolder)
            self.cfgWindow.getWidget('btn-ok').connect('clicked', lambda btn: self.cfgWindow.hide())
            self.cfgWindow.getWidget('btn-cancel').connect('clicked', lambda btn: self.cfgWindow.hide())
            self.cfgWindow.getWidget('btn-help').connect('clicked', self.onHelp)

        if not self.cfgWindow.isVisible():
            self.fillMFList()

        self.cfgWindow.show()


    def onHelp(self, btn):
        """ Display a small help message box """
        helpMsg  = _('This is the list of the folders displayed in the main window. You can add/remove folders by using the corresponding buttons')
        infoMsgBox(self.cfgWindow, helpMsg)


    def onAddMediaFolder(self, btn):
        """ Let the user add a new media folder to the list """
        if self.cfgDlgAdd is None:
            wTree              = gtk.glade.XML(os.path.join(consts.dirRes, 'FileExplorer.glade'), domain=consts.appNameShort)
            self.cfgDlgAdd     = wTree.get_widget('dlg-add')
            self.cfgDlgBtnAdd  = wTree.get_widget('btn-dlgAdd')
            self.cfgDlgTxtName = wTree.get_widget('txt-dlgName')
            self.cfgDlgTxtPath = wTree.get_widget('txt-dlgPath')
            self.cfgDlgAdd.set_title(MOD_NAME)
            self.cfgDlgAdd.set_transient_for(self.cfgWindow)
            # Connect handlers
            wTree.get_widget('btn-dlgOpen').connect('clicked', self.onDlgAddOpen)
            self.cfgDlgTxtName.connect('changed', self.onDlgAddTxtChanged)
            self.cfgDlgTxtPath.connect('changed', self.onDlgAddTxtChanged)

        self.cfgDlgBtnAdd.set_sensitive(False)
        self.cfgDlgTxtName.set_text('')
        self.cfgDlgTxtPath.set_text('')
        self.cfgDlgTxtName.grab_focus()

        if self.cfgDlgAdd.run() == gtk.RESPONSE_OK:
            name = self.cfgDlgTxtName.get_text()
            path = self.cfgDlgTxtPath.get_text()

            if name not in self.mediaFolders:
                self.mediaFolders[name] = path
                prefs.set(__name__, 'media-folders', self.mediaFolders)
                self.fillMFList()
                modules.postMsg(modules.MSG_CMD_EXPLORER_ADD, {'modName': __name__, 'expName': name, 'icon': None, 'usrParam': path, 'widget': self.scrolled})

        self.cfgDlgAdd.hide()


    def fillMFList(self):
        """ Fill the list of currently used media folders """
        self.cfgTree.freeze_child_notify()
        self.cfgStore.clear()
        if len(self.mediaFolders) == 0:
            self.cfgBtnRemove.set_sensitive(False)
        else:
            sortedMF = self.mediaFolders.items()
            sortedMF.sort()
            for mf in sortedMF:
                #self.cfgStore.append(('<b>%s</b>' % mf[0], '<i>%s</i>' % mf[1]))
                self.cfgStore.append(('%s' % mf[0], '%s' % mf[1]))
            self.cfgBtnRemove.set_sensitive(True)
            self.cfgTree.set_cursor(0)
        self.cfgTree.thaw_child_notify()


    def onDlgAddOpen(self, btn):
        """ Let the user choose a folder, and fill the corresponding field in the dialog """
        path = fileChooser.openDirectory(self.cfgDlgAdd, _('Choose a folder'))
        if path is not None:
            self.cfgDlgTxtPath.set_text(path)


    def onDlgAddTxtChanged(self, entry):
        """ Enable/disable the OK button based on the content of the text fields """
        self.cfgDlgBtnAdd.set_sensitive(self.cfgDlgTxtName.get_text() != '' and self.cfgDlgTxtPath.get_text() != '')


    def onRemoveMediaFolder(self, btn):
        """ Remove the selected media folder """
        iter = self.cfgTree.get_selection().get_selected()[1]
        if iter is None:
            self.cfgBtnRemove.set_sensitive(False)
        else:
            name = self.cfgStore.get_value(iter, 0)

            modules.postMsg(modules.MSG_CMD_EXPLORER_REMOVE, {'modName': __name__, 'expName': name})
            del self.mediaFolders[name]
            prefs.set(__name__, 'media-folders', self.mediaFolders)

            if   self.cfgStore.remove(iter): self.cfgTree.set_cursor(self.cfgStore.get_path(iter))
            elif len(self.cfgStore) != 0:    self.cfgTree.set_cursor(len(self.cfgStore)-1)
            else:                            self.cfgBtnRemove.set_sensitive(False)
