# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gobject, gtk, media, modules, os.path, pango, tools, urllib

from gtk         import gdk
from gui         import fileChooser
from tools       import consts
from media       import playlist
from gettext     import gettext as _
from modules     import Module
from media.track import Track


(
    ROW_PTH,  # Full path to the file
    ROW_NUM,  # Track number
    ROW_TIT,  # Track title
    ROW_ART,  # Track artist
    ROW_ALB,  # Track album
    ROW_LEN,  # Track length in seconds
    ROW_ICO   # Icon drawn in front of the row
) = range(7)


class Tracklist(Module):
    """
        This module manages the tracklist
    """


    def __init__(self, wTree):
        """ Constructor """
        Module.__init__(self)
        self.index    = None
        self.playtime = 0
        # Drag and drop
        self.mayDrag = False
        self.context = None
        # Retrieve widgets
        self.tree      = wTree.get_widget('tree-tracklist')
        self.window    = wTree.get_widget('win-main')
        self.btnClear  = wTree.get_widget('btn-tracklistClear')
        self.popupTree = None
        self.tracklist = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_STRING,
                                       gobject.TYPE_STRING, gobject.TYPE_INT, gdk.Pixbuf)
        self.tree.set_model(self.tracklist)
        self.btnClear.set_sensitive(False)
        # The renderers used for the tree
        txtRightRdr = gtk.CellRendererText()
        txtLeftRdr  = gtk.CellRendererText()
        txtRightRdr.set_property('xalign', 1.0)
#        txtLeftRdr.set_property('ellipsize', pango.ELLIPSIZE_END)
        # The icons used in the tree
        self.nullIcon  = gdk.Pixbuf(gdk.COLORSPACE_RGB, True, 8, 16, 16)
        self.playIcon  = self.tree.render_icon(gtk.STOCK_MEDIA_PLAY,  gtk.ICON_SIZE_MENU)
        self.pauseIcon = self.tree.render_icon(gtk.STOCK_MEDIA_PAUSE, gtk.ICON_SIZE_MENU)
        self.nullIcon.fill(0x00000000)
        # Create the columns of the tree
        self.tree.insert_column_with_attributes(-1, '#', gtk.CellRendererPixbuf(), pixbuf=ROW_ICO)
        self.tree.get_column(0).pack_start(txtRightRdr, True)
        self.tree.get_column(0).add_attribute(txtRightRdr, 'text', ROW_NUM)
        self.tree.get_column(0).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)

        self.tree.insert_column_with_attributes(-1, _('Title'),  txtLeftRdr, text=ROW_TIT)
        self.tree.get_column(1).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(1).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Artist'), txtLeftRdr, text=ROW_ART)
        self.tree.get_column(2).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(2).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Album'),  txtLeftRdr, text=ROW_ALB)
        self.tree.get_column(3).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(3).set_expand(True)

        self.tree.insert_column_with_attributes(-1, _('Length'), txtRightRdr, text=ROW_LEN)
        self.tree.get_column(4).set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
        self.tree.get_column(4).set_expand(False)

        self.tree.get_column(4).set_cell_data_func(txtRightRdr, lambda c, cll, mdl, it: cll.set_property('text', tools.sec2str(mdl.get_value(it, ROW_LEN))))
        dndTargets = [('text/uri-list', 0, 0), ('internal', gtk.TARGET_SAME_WIDGET, 0)]
        self.tree.drag_source_set(gdk.BUTTON1_MASK, dndTargets, gdk.ACTION_DEFAULT)
        self.tree.enable_model_drag_dest(dndTargets, gdk.ACTION_DEFAULT)
        self.tree.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        # Messages handled by this module
        modules.register(self, [modules.MSG_CMD_TRACKLIST_SET, modules.MSG_CMD_TRACKLIST_CLR, modules.MSG_CMD_TRACKLIST_ADD,
                                modules.MSG_CMD_TOGGLE_PAUSE, modules.MSG_CMD_NEXT, modules.MSG_CMD_PREVIOUS, modules.MSG_EVT_NEED_BUFFER,
                                modules.MSG_EVT_STOPPED, modules.MSG_EVT_PAUSED, modules.MSG_EVT_UNPAUSED, modules.MSG_EVT_TRACK_ENDED])
        # GTK handlers
        self.tree.connect('drag-begin',           self.onDragBegin)
        self.tree.connect('key-press-event',      self.onKeyboard)
        self.tree.connect('button-press-event',   self.onButtonPress)
        self.tree.connect('drag-data-received',   self.onDragDataReceived)
        self.tree.connect('button-release-event', self.onButtonRelease)
        self.btnClear.connect('clicked', lambda widget: modules.postMsg(modules.MSG_CMD_TRACKLIST_CLR))


    def hasNext(self):      return self.index is not None and self.index < len(self.tracklist) - 1
    def hasPrev(self):      return self.index is not None and self.index > 0
    def getAllFiles(self):  return [row[ROW_PTH] for row in self.tracklist]


    def setVisible(self, widget, show):
        """ Show/hide a widget """
        if show: widget.show()
        else:    widget.hide()


    def jumpTo(self, index):
        """ Jump to the track located at the given index """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.nullIcon)
        self.index = index
        self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.playIcon)
        self.tree.scroll_to_cell(self.index)
        modules.postMsg(modules.MSG_CMD_PLAY, {'filename': self.tracklist.get_value(self.tracklist.get_iter(self.index), ROW_PTH)})
        modules.postMsg(modules.MSG_EVT_NEW_TRACK, {'track': Track(self.tracklist.get(self.tracklist.get_iter(self.index), 0, 1, 2, 3, 4, 5))})
        modules.postMsg(modules.MSG_EVT_TRACK_MOVED, {'hasPrevious': self.hasPrev(), 'hasNext': self.hasNext()})


    def onTrackEnded(self):
        """ The current track has ended, jump to the next one if any """
        if self.hasNext(): self.jumpTo(self.index+1)
        else:              modules.postMsg(modules.MSG_CMD_STOP)


    def onBufferingNeeded(self):
        """ The current track is close to its end, se we try to buffer the next one to avoid gaps """
        if self.hasNext():
            modules.postMsg(modules.MSG_CMD_BUFFER, {'filename': self.tracklist.get_value(self.tracklist.get_iter(self.index+1), ROW_PTH)})


    def insert(self, position, files):
        """ Insert some files in the tracklist, append them if position is None """
        tracks = media.getTracks(files)
        if len(tracks) == 0:
            return
        # Update the current index if needed
        if self.index is not None and position is not None and position <= self.index:
            self.index += len(tracks)
        # Insert each track in the tree
        self.tree.freeze_child_notify()
        for track in tracks:
            track.append(self.nullIcon)
            self.playtime += track[ROW_LEN]
            if position is None:
                self.tracklist.append(track)
            else:
                self.tracklist.insert(position, track)
                position += 1
        self.tree.thaw_child_notify()
        self.btnClear.set_sensitive(True)
        modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': self.getAllFiles(), 'playtime': self.playtime})
        modules.postMsg(modules.MSG_EVT_TRACK_MOVED, {'hasPrevious': self.hasPrev(), 'hasNext': self.hasNext()})


    def set(self, files, playNow):
        """ Replace the tracklist, clear it if files is None """
        self.playtime = 0
        self.index    = None
        self.tracklist.clear()

        if files is None:
            modules.postMsg(modules.MSG_CMD_STOP)
            self.btnClear.set_sensitive(False)
            modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': [], 'playtime': 0})
        else:
            self.insert(None, files)
            if playNow:
                self.jumpTo(0)


    def onStopped(self):
        """ Playback has been stopped """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, self.nullIcon)
            self.index = None


    def onPausedToggled(self, icon):
        """ Switch between paused and unpaused """
        if self.index is not None:
            self.tracklist.set_value(self.tracklist.get_iter(self.index), ROW_ICO, icon)


    def playSelection(self):
        """ Start playing the currently selected track """
        if self.tree.get_selection().count_selected_rows() == 1:
            self.jumpTo(self.tree.get_selection().get_selected_rows()[1][0][0])


    def savePlaylist(self):
        """ Save the current tracklist to a playlist """
        directory = os.path.dirname(self.tracklist.get_value(self.tracklist.get_iter(0), ROW_PTH))
        filename  = fileChooser.save(self.window, _('Save playlist'), 'playlist.m3u', directory)
        if filename is not None:
            playlist.save(self.getAllFiles(), filename)


    def cropSelection(self):
        """ Remove all tracks but the selection """
        selection = self.tree.get_selection()
        pathsList = selection.get_selected_rows()[1]
        self.tree.freeze_child_notify()
        selection.select_all()
        for path in pathsList:
            selection.unselect_path(path[0])
        self.removeSelection()
        selection.select_all()
        self.tree.thaw_child_notify()


    def removeSelection(self):
        """ Remove all the currently selected tracks """
        # Keep an eye on the current track, if any
        if self.index is None: currentIter = None
        else:                  currentIter = self.tracklist.get_iter(self.index)

        # Remove all the selected tracks and move focus accordingly
        self.tree.freeze_child_notify()
        for iter in [self.tracklist.get_iter(path[0]) for path in self.tree.get_selection().get_selected_rows()[1]]:
            if self.tracklist.get_value(iter, ROW_ICO) != self.nullIcon:
                modules.postMsg(modules.MSG_CMD_STOP)
                self.index  = None
                currentIter = None

            self.playtime -= self.tracklist.get_value(iter, ROW_LEN)

            if   self.tracklist.remove(iter): self.tree.set_cursor(self.tracklist.get_path(iter))
            elif len(self.tracklist) != 0:    self.tree.set_cursor(len(self.tracklist)-1)
            else:                             self.btnClear.set_sensitive(False)
        self.tree.thaw_child_notify()

        # Retrieve the new index if needed
        if currentIter is not None:
            self.index = self.tracklist.get_path(currentIter)[0]

        modules.postMsg(modules.MSG_EVT_TRACK_MOVED,   {'hasPrevious': self.hasPrev(),     'hasNext':  self.hasNext()})
        modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files':       self.getAllFiles(), 'playtime': self.playtime})


    def showPopupMenu(self, path, button, time):
        """ The index parameter may be None """
        if self.popupTree is None:
            self.popupTree = gtk.glade.XML(os.path.join(consts.dirRes, 'TracklistMenu.glade'), domain=consts.appNameShort)
            # Connect signals
            self.popupTree.get_widget('item-save').connect('activate', lambda wdg: self.savePlaylist())
            self.popupTree.get_widget('item-play').connect('activate', lambda wdg: self.playSelection())
            self.popupTree.get_widget('item-crop').connect('activate', lambda wdg: self.cropSelection())
            self.popupTree.get_widget('item-stop').connect('activate', lambda wdg: modules.postMsg(modules.MSG_CMD_STOP))
            self.popupTree.get_widget('item-pause').connect('activate', lambda wdg: modules.postMsg(modules.MSG_CMD_TOGGLE_PAUSE))
            self.popupTree.get_widget('item-clear').connect('activate', lambda wdg: modules.postMsg(modules.MSG_CMD_TRACKLIST_CLR))
            self.popupTree.get_widget('item-remove').connect('activate', lambda wdg: self.removeSelection())
            self.popupTree.get_widget('item-unpause').connect('activate', lambda wdg: modules.postMsg(modules.MSG_CMD_TOGGLE_PAUSE))

        # Enable only relevant menu entries
        if path is None: icon = None
        else:            icon = self.tracklist.get_value(self.tracklist.get_iter(path), ROW_ICO)

        nbRows = self.tree.get_selection().count_selected_rows()

        if nbRows == 1:
            self.setVisible(self.popupTree.get_widget('item-play'), icon == self.nullIcon)
            self.popupTree.get_widget('item-play').set_sensitive(icon == self.nullIcon)

            self.setVisible(self.popupTree.get_widget('item-unpause'), icon == self.pauseIcon)
            self.popupTree.get_widget('item-unpause').set_sensitive(icon == self.pauseIcon)

            self.setVisible(self.popupTree.get_widget('item-pause'), icon == self.playIcon)
            self.popupTree.get_widget('item-pause').set_sensitive(icon == self.playIcon)

            self.setVisible(self.popupTree.get_widget('item-stop'), (icon == self.playIcon or icon == self.pauseIcon))
            self.popupTree.get_widget('item-stop').set_sensitive(icon == self.playIcon or icon == self.pauseIcon)
        else:
            self.popupTree.get_widget('item-play').show()
            self.popupTree.get_widget('item-play').set_sensitive(False)
            self.popupTree.get_widget('item-unpause').hide()
            self.popupTree.get_widget('item-pause').hide()
            self.popupTree.get_widget('item-stop').hide()

        self.popupTree.get_widget('item-sep1').set_sensitive(icon is not None)
        self.popupTree.get_widget('item-crop').set_sensitive(icon is not None)
        self.popupTree.get_widget('item-remove').set_sensitive(icon is not None)
        self.popupTree.get_widget('item-sep2').set_sensitive(icon is not None)

        self.popupTree.get_widget('menu-popup').popup(None, None, None, button, time)


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ A message has been received """
        if   msg == modules.MSG_EVT_PAUSED:                              self.onPausedToggled(self.pauseIcon)
        elif msg == modules.MSG_EVT_STOPPED:                             self.onStopped()
        elif msg == modules.MSG_EVT_UNPAUSED:                            self.onPausedToggled(self.playIcon)
        elif msg == modules.MSG_EVT_TRACK_ENDED:                         self.onTrackEnded()
        elif msg == modules.MSG_EVT_NEED_BUFFER:                         self.onBufferingNeeded()
        elif msg == modules.MSG_CMD_TRACKLIST_SET:                       self.set(params['files'], params['playNow'])
        elif msg == modules.MSG_CMD_TRACKLIST_ADD:                       self.insert(None, params['files'])
        elif msg == modules.MSG_CMD_TRACKLIST_CLR:                       self.set(None, False)
        elif msg == modules.MSG_CMD_NEXT and self.hasNext():             self.jumpTo(self.index+1)
        elif msg == modules.MSG_CMD_PREVIOUS and self.hasPrev():         self.jumpTo(self.index-1)
        elif msg == modules.MSG_CMD_TOGGLE_PAUSE and self.index is None: self.jumpTo(0)


    # --== GTK handlers ==--


    def onButtonRelease(self, tree, event):
        """ We need this handler to handle D'n'Drop with multiple selections """
        path = self.tree.get_path_at_pos(int(event.x), int(event.y))
        if self.mayDrag and path is not None:
            self.tree.grab_focus()
            self.tree.get_selection().unselect_all()
            self.tree.get_selection().select_path(path[0][0])
        elif self.context is not None:
            self.onDragDrop(tree, self.context, gtk.get_current_event().x, gtk.get_current_event().y, event.time)
            self.context = None
        self.mayDrag = False


    def onButtonPress(self, tree, event):
        """ Play the selected track on double click, or show a popup menu on right click """
        pathInfo  = self.tree.get_path_at_pos(int(event.x), int(event.y))
        selection = self.tree.get_selection()

        if pathInfo is None: path = None
        else:                path = pathInfo[0][0]

        # Left click
        if event.button == 1:
            if path is None:
                selection.unselect_all()
            elif event.type == gdk._2BUTTON_PRESS:
                self.jumpTo(path)
            elif event.state == 0 and selection.count_selected_rows() > 1 and selection.path_is_selected(path):
                self.mayDrag = True
                return True
        # Right click
        elif event.button == 3:
            if path is None:
                selection.unselect_all()
            elif not selection.path_is_selected(path):
                selection.unselect_all()
                selection.select_path(path)

            if len(self.tracklist) != 0:
                self.showPopupMenu(path, event.button, event.time)
                return True


    def onKeyboard(self, tree, event):
        """ Remove the selection if possible """
        if gdk.keyval_name(event.keyval) == 'Delete':
            self.removeSelection()


    def onDragBegin(self, tree, context):
        """ Drag'n'Drop begins """
        self.mayDrag = False
        self.context = context


    def onDragDataReceived(self, tree, context, x, y, dragData, info, time):
        """ External Drag'n'Drop """
        files    = [urllib.url2pathname(uri.strip()[7:]) for uri in dragData.data.split('\n') if uri.startswith('file://')]
        dropInfo = tree.get_dest_row_at_pos(int(x), int(y))

        if len(files) == 0:
            context.finish(False, False, time)

        if dropInfo is None: self.insert(None, files)
        else:                self.insert(dropInfo[0][0], files)

        context.finish(True, False, time)


    def onDragDrop(self, tree, context, x, y, time):
        """ Internal Drag'n'Drop """
        dropInfo = tree.get_dest_row_at_pos(int(x), int(y))
        iterList = [self.tracklist.get_iter(path[0]) for path in self.tree.get_selection().get_selected_rows()[1]]

        if dropInfo is not None:
            path, pos  = dropInfo
            targetIter = self.tracklist.get_iter(path)
            if pos == gtk.TREE_VIEW_DROP_AFTER or pos == gtk.TREE_VIEW_DROP_INTO_OR_AFTER:
                iterList.reverse()

        self.tree.freeze_child_notify()
        for srcIter in iterList:
            srcIndex = self.tracklist.get_path(srcIter)[0]
            # Get an iterator on the destination
            if dropInfo is None:
                dstIter = self.tracklist.append(self.tracklist[srcIter])
            elif pos == gtk.TREE_VIEW_DROP_BEFORE or pos == gtk.TREE_VIEW_DROP_INTO_OR_BEFORE:
                dstIter = self.tracklist.insert_before(targetIter, self.tracklist[srcIter])
            else:
                dstIter = self.tracklist.insert_after(targetIter, self.tracklist[srcIter])
            # Remove the source row(s)
            self.tracklist.remove(srcIter)
            dstIndex = self.tracklist.get_path(dstIter)[0]
            # Move the current track index accordingly
            if   srcIndex == self.index:                           self.index = dstIndex
            elif srcIndex < self.index and dstIndex >= self.index: self.index -= 1
            elif srcIndex > self.index and dstIndex <= self.index: self.index += 1
        self.tree.thaw_child_notify()

        modules.postMsg(modules.MSG_EVT_NEW_TRACKLIST, {'files': self.getAllFiles(),   'playtime': self.playtime})
        modules.postMsg(modules.MSG_EVT_TRACK_MOVED,   {'hasPrevious': self.hasPrev(), 'hasNext':  self.hasNext()})
